#include <stdio.h>
#include <dir.h>
#include <alloc.h>

#define szficdef     18
#define szbuf     16384
#define szlistini    64

#define octet char


/* nom d'un fichier + nature */
typedef char ficdef [szficdef];

/* buffer de transfert */
static octet buffer [szbuf];

/* indicateur : 1 = on recopie avec l'ordre du rpertoire d'origine */
static octet opto = 0;

/* nombre de Koctets, de fichiers, de rpertoires transfrs */
static long nblocs = 0;
static int  nfic = 0, ndir = 0;

/* date et heure au lancement de la commande */
static unsigned long hdebut;

/* date de dernire mise  jour du fichier  recopier */
static unsigned long mtime;

/* indicateur d'absence d'erreur pendant la copie */
static octet recopie = 1;

/* numro du disque de destination */
static char disque;

/* descripteurs de fichiers (pour la fonction fopen) */
static FILE *ficin, *ficout;

/* descripteur de fichier (pour la fonction open) */
static int descfic;

/* taille des zones de donnes lues et recopies */
static int szlect, szrecop;

/* chemin d'accs aux fichiers du rpertoire selectionn */
char chemin [256];


/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    /* permet de dterminer le remplissage du disque pendant la copie */
    unsigned place [5];

    /* rcupration de l'option -o ventuelle */
    if (narg > 1 && strcmp (varg [1], "-o") == 0)
    {
        opto = 1;
        narg --;
        varg ++;
    }

    /* test du nombre d'arguments */
    if (narg != 3)
        fprintf (stderr, "%s\n",
        "Syntaxe : copie [-o] <rpertoire d'origine> <rpertoire de destination>");
    else
    {
        /* vrifier que les 2 arguments sont des rpertoires */
        testchemin (varg [1]);
        testchemin (varg [2]);

        /* rcuprer la date et l'heure courantes */
        recuptime ();

        /* rcuprer le nom du disque destinataire de la copie */
        if (varg [2][1] == ':')
            disque = *varg [2] & 0x1f;
        else
            disque = 0;    /* disque courant */

        /* mesurer la place libre sur le disque avant transfert */
        getdfree (disque, place);

        /* copie des fichiers */
        copdir (varg [1], varg [2]);

        /* mesurer la place libre sur le disque aprs transfert */
        getdfree (disque, &place [1]);

        /* calcul de nombre de Koctets utiliss */
        nblocs = ((long) place [0] - (long) place [1]) *
                       ((place [3] * place [4]) / 512);

        /* rcapitulatif des transferts */
        if (nfic == 1)
            printf ("  1 fichier transfr  ");
        else
            printf ("%3d fichiers transfrs", nfic);


        if (nblocs == 0)
            printf ("    Occupation disque inchange");

        else if (nblocs > 2)
            printf ("  %3ld Koctets utiliss", (nblocs + 1) / 2);

        else if (nblocs < -2)
            printf ("  %3ld Koctets librs", (1 - nblocs) / 2);

        else if (nblocs > 0)
            printf ("    1 Koctet utilis  ");
        else
            printf ("    1 Koctet libr ");


        if (ndir == 1)
            printf ("  1 rpertoire cr");
        else if (ndir > 1)
            printf (" %2d rpertoires crs", ndir);

        putchar (10);
    }
}


/* mettre "chaine" sous forme normalise et vrifier
   que c'est un chemin d'accs  un rpertoire      */
testchemin (chaine)
char chaine [];
{
    /* table de description d'un fichier */
    struct ffblk fich;

    /* descripteur de fichier */
    FILE *ftest;

    /* caractre de "chaine" sur lequel on est positionn */
    char *poschn;


    /* positionnement sur le premier caractre de "chaine" */
    poschn = chaine;

    /* rpter */
    do
    {
        /* si le caractre courant est un '\' */
        if (*poschn == '\\')
            /* le remplacer par '/' */
            *poschn = '/';

        /* sinon si le caractre est une lettre minuscule */
        else if (*poschn >= 'a')
            /* le remplacer par la lettre majuscule correspondante */
            *poschn = *poschn - 0x20;

        /* passer au caractre suivant */
        poschn++;
    }
    /* jusque toute la chaine a t explore */
    while (*poschn != 0);

    /* si le dernier caractre de la chaine est un '/' */
    if (*(poschn - 1) == '/')
        /* supprimer ce caractre */
        *(poschn - 1) = 0;

    /* fabriquer le chemin d'accs aux lments du rpertoire  tester */
    sprintf (chemin, "%s/*.*", chaine);

    /* si on n'a pas trouv de fichier dans le rpertoire */
    if (findfirst (chemin, &fich, 0x10) != 0)
    {
        /* vrification supplmentaire */
        sprintf (chemin, "%s/-", chaine);

        /* essayer de crer un fichier dans le rpertoire */
        ftest = fopen (chemin, "a");

        /* si l'essai est concluant */
        if (ftest > 0)
        {
            /* fermer et dtruire le fichier d'essai */
            fclose (ftest);
            unlink (chemin);
        }
        /* sinon */
        else
        {
            /* message d'erreur */
            fprintf (stderr, "%s n'est pas un rpertoire\n", chaine);

            /* sortir du programme */
            exit ();
        }
    }
}


/* rcupration de la date et de l'heure courante */
recuptime ()
{
    unsigned char heure [4];
    unsigned int jour [2];

    /* rcupration de la date et de l'heure */
    getdate (jour);
    gettime (heure);

    /* conversion dans le format employ pour les fichiers */
    hdebut = ((jour [0] - 1980) << 9) | ((jour [1] & 0xff00) >> 3)
                                               | (jour [1] & 0xff);
    hdebut = (hdebut << 16) | (((heure [1] << 11) | (heure [0] << 5)
                              | (heure [3] >> 1)) & 0xffff);
}


/* copie du rpertoire "orig" et de son
   arborescence dans le rpertoire "dest" */
copdir (orig, dest)
char orig [], dest [];
{
    /* chemin d'accs aux fichiers des diffrents rpertoires */
    char *ficorig, *ficdest;

    /* table de description d'un fichier */
    struct ffblk fich;

    /* indicateur */
    int trouve;

    /* compteurs */
    int posliste, nbfic, szliste;

    /* liste des noms et attribut des fichiers d'un rpertoire */
    ficdef *listefic, *newliste;


    /* allouer les chaines ficorig et ficdest avec la bonne taille mmoire */
    ficorig = malloc (strlen (orig) + szficdef);
    ficdest = malloc (strlen (dest) + szficdef);

    /* allocation du tableau listefic */
    if (opto == 0)
        szliste = szlistini;
    else
        szliste = 1;

    listefic = malloc (szliste * szficdef);

    if (listefic == 0 || ficdest == 0)
    {
        if (opto == 0)
            fprintf (stderr,
    "Manque de place mmoire, utiliser l'option -o ou continuer avec copnew\n");
        else
            fprintf (stderr, "Grave problme de gestion mmoire\n");

        exit ();
    }
    
    /* fabriquer le chemin d'accs aux lments du rpertoire  recopier */
    sprintf (chemin, "%s/*.*", orig);

    /* recherche du premier fichier de ce rpertoire
       on exclue un ventuel fichier label de la recherche */
    trouve = findfirst (chemin, &fich, 0xf7);

    /* si ce fichier est "." */
    if (strcmp (fich.ff_name, ".") == 0)
    {
        /* passer au fichier suivant */
        trouve = findnext (&fich);

        /* si ce fichier est ".." */
        if (strcmp (fich.ff_name, "..") == 0)
            /* passer au fichier suivant */
            trouve = findnext (&fich);
    }

    /* initialisation : nombre de fichiers mmoriss */
    nbfic = 0;

    /* tantque on trouve des nouveaux fichiers dans le rpertoire */
    while (trouve == 0 && recopie > 0)
    {
        /* si tri alphabtique des fichiers */
        if (opto == 0)
        {
            /* si liste de fichiers trop petite */
            if (nbfic >= szliste)
            {
                newliste = realloc (listefic, (szliste + szlistini) * szficdef);

                /* prise en compte du rsultat de l'opration */
                if (newliste != 0)
                {
                    /* nouvelle taille et nouvelle adresse de la liste */
                    szliste  = szliste + szlistini;
                    listefic = newliste;
                }
                else
                {
                    /* dbordement, on copie le premier fichier class */
                    sprintf (ficorig, "%s/%s", orig, *listefic);
                    sprintf (ficdest, "%s/%s", dest, *listefic);

                    copfic (*listefic, ficorig, ficdest);

                    /* et on le supprime de la liste */
                    memmove (listefic [0], listefic [1], --nbfic * szficdef);
                }
            }

            /* chercher l'lment de la liste o le nouveau fichier
               doit tre insr pour respecter l'ordre alphabtique */
            for (posliste = 0; (posliste < nbfic) &&
                (strcmp (listefic [posliste], fich.ff_name)) < 0; posliste++);

            /* si la liste n'a pas t parcourrue en entier */
            if (posliste < nbfic)
                /* dcaler les fichiers restants d'une position */
                memmove (listefic [posliste + 1],
                        listefic [posliste], (nbfic - posliste) * szficdef);

            /* incrmenter le nombre de fichiers mmoriss */
            nbfic++;
        }
        /* sinon, la liste se limite  un fichier */
        else
            posliste = 0;

        /* recopier le nom du fichier dans la liste */
        memmove (listefic [posliste], fich.ff_name, 13);

        /* recopier l'attribut du fichier */
        listefic [posliste][13] = fich.ff_attrib;

        /* recopier la date de mise a jour du fichier */
        memmove (&listefic [posliste][14], &fich.ff_ftime, 4);

        if (opto == 1)
        {
            sprintf (ficorig, "%s/%s", orig, *listefic);
            sprintf (ficdest, "%s/%s", dest, *listefic);

            copfic (*listefic, ficorig, ficdest);
        }

        /* rechercher le fichier suivant du rpertoire */
        trouve = findnext (&fich);
    }

    /* recopier tous les fichiers mmoriss */
    for (posliste = 0; posliste < nbfic && recopie > 0; posliste++)
    {
        sprintf (ficorig, "%s/%s", orig, listefic [posliste]);
        sprintf (ficdest, "%s/%s", dest, listefic [posliste]);

        copfic (listefic [posliste], ficorig, ficdest);
    }

    /* librer l'espace mmoire allou */
    free (listefic);
    free (ficorig);
    free (ficdest);
}


/* recopie d'un fichier slectionn */
copfic (ficdef fichier, char *ficorig, char *ficdest)
{
    /* place restante sur le disque */
    int place [4];


    /* si le fichier  recopier peut tre lu */
    if (access (ficorig, 4) == 0)
    {
        /* si le fichier  recopier est un rpertoire */
        if (fichier [13] & 0x10)
        {
            /* rcuprer la date de cration du rpertoire */
            memcpy (&mtime, &fichier [14], 4);

            /* si ce rpertoire n'a pas t cr pendant la copie */
            if (mtime < hdebut)
            {
                /* diter son nom */
                puts (ficdest);

                /* si le rpertoire de destination n'existe pas */
                if (access (ficdest, 0) == -1)
                {
                    /* la crer */
                    mkdir (ficdest);
                    ndir++;
                }

                /* recopier l'arborescence du rpertoire */
                copdir (ficorig, ficdest);
            }
        }
        /* sinon */
        else
        {
            /* diter le nom du fichier destination */
            puts (ficdest);

            /* ouvrir les fichiers origine et destination */
            ficin  = fopen (ficorig, "rb");
            ficout = fopen (ficdest, "wb");

            /* lire un morceau du fichier origine */
            szlect = fread (buffer, 1, szbuf, ficin);

            /* tantque non fin de fichier */
            while ((szlect > 0) && (recopie > 0))
            {
                /* recopier la zone lue */
                szrecop = fwrite (buffer, 1, szlect, ficout);

                /* si problme lors de la copie */
                if (szrecop < szlect)
                {
                    /* on arrte la copie */
                    recopie = 0;

                    /* manque d'espace disque ? */
                    getdfree (disque, place);

                    /* message d'erreur */
                    if ((place [0] * place [2] * place [3]) < szlect)
                        fprintf (stderr, "Manque d'espace disque\n");
                    else
                        fprintf (stderr, "Erreur indtermine\n");
                }
                /* sinon */
                else
                    /* lire la suite du fichier d'origine */
                    szlect = fread (buffer, 1, szbuf, ficin);
            }

            /* fermer les fichiers */
            fclose (ficin);
            fclose (ficout);

            /* si la copie s'est bien passe */
            if (recopie > 0)
            {
                /* rcuprer la date de mise  jour du fichier d'origine */
                memcpy (&mtime, &fichier [14], 4);

                /* corriger la date de mise  jour du fichier destination */
                descfic = open (ficdest, 0x8001);
                setftime (descfic, &mtime);
                close (descfic);

                nfic++;
            }
            /* sinon dtruire le fichier mal recopi */
            else
                unlink (ficdest);
        }
    }
    /* sinon (le fichier ne peut pas tre lu */
    else
        /* message d'erreur */
        fprintf (stderr, "%s : Droit d'accs insuffisant\n", ficorig);
}
