/* Compare 2 fichiers */

#include <stdio.h>
#include <alloc.h>

#define SZMAXBUF  65520     /* taille maximale des buffers caractre */
#define SZMAXLIG    256     /* longueur maximale des lignes avant troncation */
#define MAXDIF     1000     /* nombre max de lignes diffrentes conscutives */
#define RESERVEMEM 2000     /* mmoire systme  ne pas allouer explicitement */

#define TAB   9
#define CTRZ 26


#define octet unsigned char
#define int16 unsigned short


typedef struct ficinfo
{
    FILE *fic;              /* descripteur du fichier */
    char far *buffer;       /* mmorise les lignes du fichier */
    int16 nbmemo;           /* nombre de lignes memorises */
    int16 depl;             /* nombre ligne diffrentes avant synchro */
    long  numl;             /* numro de la ligne courante */
    octet finfic;           /* indique si fin de fichier rencontre */
    octet szlig  [MAXDIF];  /* taille des lignes memorises */
    int16 deblig [MAXDIF];  /* dbut des lignes dans buffer  */
};

typedef struct ligne
{
    char far *caract;
    octet szlig;
    octet poslig;
};


static octet optb = 0;      /* slection de l'option -b */
static octet optd = 0;      /* slection de l'option -d */
static octet optt = 0;      /* slection de l'option -t */

static int16 szbuf;         /* taille effective des buffers caractre */

static struct ficinfo fichier [2];
static struct ficinfo *ficur;  /* fichier courant (l'un des 2 prcdents) */

static struct ligne ligne [2], *ligcur; /* utilise dans eqlig et sautblanc */

main (narg, varg)
int narg;
char *varg [];
{
    octet i;


    /* se positionner sur le premier argument */
    varg++;
    narg--;

    /* rcupration des options -b et -t */
    while (**varg == '-')
    {
        if (strcmp (*varg, "-b") == 0)
            optb = 1;

        else if (strcmp (*varg, "-d") == 0)
            optd = 1;

        else if (strcmp (*varg, "-t") == 0)
            optt = 1;

        else
        {
            /* option inconnue */
            psyntaxe ();
            exit ();
        }

        /* passage  l'argument suivant */
        varg++;
        narg--;
    }

    /* si nombre d'arguments correct */
    if (narg == 2)
    {
        /* prendre en compte l'option -d */
        if (optd)
        {
            /* rajouter le nom de fichier  l'argument 1 */
            *(varg [1] - 1) = '/';

            /* supprimer le / en trop si ncessaire */
            if (*(varg [1] - 2) == '/' || *(varg [1] - 2) == '\\')
            {
                strcpy (varg [1] - 1, varg [1]);
                varg [1] = varg [1] - 1;
            }
        }

        for (i = 0; i < 2; i++)
        {
            /* fichier 1 ou 2 */
            ficur = &fichier [i];

            /* ouvrir un fichier */
            ficur->fic = fopen (varg [i], "rb");

            if (ficur->fic == 0)
            {
                fprintf (stderr, "%s : inexistant\n", varg [i]);
                exit ();
            }
        }

        /* allouer les buffers mmoire */
        if ((farcoreleft () / 2) > (long) SZMAXBUF + RESERVEMEM)
            szbuf = SZMAXBUF;
        else
            szbuf = (farcoreleft () - RESERVEMEM) / 2;

        fichier[0].buffer = farmalloc (szbuf);
        fichier[1].buffer = farmalloc (szbuf);

        if (fichier [1].buffer == 0)
        {
            fputs ("Manque de place mmoire\n", stderr);
            exit ();
        }

        /* comparer les fichiers */
        cmpfic ();
    }
    /* message d'erreur sur le nombre d'arguments */
    else
        psyntaxe ();
}

/* dition de la syntaxe de la commande */
psyntaxe ()
{
    fputs ("Syntaxe : diff [-b] [-t] <fichier 1> <fichier 2>\n", stderr);
    fputs ("     ou : diff [-b] [-t] -d <rpertoire> <nom de fichier>\n", stderr);
}



/* comparaison des fichiers */
cmpfic ()
{
    int16 i, j;
    octet trouve;


    /* initialisation */
    for (i = 0; i < 2; i++)
    {
        ficur = &fichier [i];

        ficur->numl = 1;
        ficur->nbmemo = 0;
        ficur->finfic = 0;
    }

    while (fichier [0].finfic == 0 && fichier [1].finfic == 0)
    {
        while (eqlig (0, 0))
        {
            if (fichier [0].finfic && fichier [1].finfic)
                return;

            supligmemo ();
        }

        i = 1;
        trouve = 0;

        do
        {
            trouve = eqlig (i, i);
            j = 0;

            while (trouve == 0 && j < i)
            {
                j++;

                trouve = eqlig (i, i - j);

                if (trouve == 0)
                    trouve = eqlig (i - j, i);
            }

            i++;
        }
        while ((trouve == 0 || resync () == 0) && i < MAXDIF);

        if (i < MAXDIF)
        {
            affdif ();
            supligmemo ();
        }
        else
        {
            puts ("Trop de diffrences : resynchronisation impossible");
            fichier [0].depl = i;
            fichier [1].depl = i;
            affdif ();
            return;
        }
    }

    if (fichier [0].finfic == 0)
    {
        printf ("%ld,-\n", fichier [0].numl);
        affin (0, '<');
    }
    else if (fichier [1].finfic == 0)
    {
        printf ("-,%ld\n", fichier [1].numl);
        affin (1, '>');
    }
}

/* comparaison d'une ligne de chaque fichier */
int eqlig (depl1, depl2)
int16 depl1;
int16 depl2;
{
    octet i;


    /* sauvegarde des dplacements */
    fichier [0].depl = depl1;
    fichier [1].depl = depl2;

    /* lecture dans les fichiers si ncessaire */
    for (i = 0; i < 2; i++)
    {
        ficur = &fichier [i];

        if (ficur->nbmemo <= ficur->depl && ficur->finfic == 0)
            lecture ();
    }

    /* si fin de fichier, pas besoin de comparer des lignes */
    if (fichier [0].nbmemo <= depl1 || fichier [1].nbmemo <= depl2)
        return (fichier [0].nbmemo <= depl1 && fichier [1].nbmemo <= depl2);

    /* cas gnral : initialisation */
    for (i = 0; i < 2; i++)
    {
        ficur  = &fichier [i];
        ligcur = &ligne [i];

        ligcur->caract = &ficur->buffer [ficur->deblig [ficur->depl]];
        ligcur->szlig  =  ficur->szlig  [ficur->depl];
        ligcur->poslig = 0;
    }

    /* on ignore les blancs en dbut de ligne si ncessaire */
    if (optb == 1)
        sautblancs ();

    /* comparaison des lignes */
    while (ligne [0].poslig < ligne [0].szlig
        && ligne [1].poslig < ligne [1].szlig)
    {
        /* termin si difference trouve */
        if (ligne [0].caract [ligne [0].poslig]
         != ligne [1].caract [ligne [1].poslig])
            return (0);

        /* sinon passer aux caractres suivants */
        else
        {
            ligne [0].poslig ++;
            ligne [1].poslig ++;

            /* si option -b, on saute les blancs */
            if (optb == 1)
                sautblancs ();
        }
    }

    /* controle suplmentaire en fin de ligne */
    if (ligne [0].poslig < ligne [0].szlig
     || ligne [1].poslig < ligne [1].szlig)
        return (0);
    else
        return (1);
}

/* saute les caractres blancs en position courante */
sautblancs ()
{
    octet i;


    for (i = 0; i < 2; i++)
    {
        ligcur = &ligne [i];

        while (ligcur->caract [ligcur->poslig] == ' '
                            && ligcur->poslig < ligcur->szlig)
            ligcur->poslig ++;
    }
}



/* dtermine si on s'est correctement resynchronis aprs une diffrence */
int resync ()
{
    int16 decalage;


    if (fichier [0].depl > fichier [1].depl)
        decalage = fichier [0].depl;
    else
        decalage = fichier [1].depl;


    switch (decalage)
    {
        case 0 :
        case 1 : return (1);

        case 2 :
        case 3 :
        case 4 : if (eqsuiv (1) || (eqsuiv (2) && eqsuiv (3)))
                     return (1);
                 else
                     return (0);

        default: if (eqsuiv (1) && (eqsuiv (2) || (eqsuiv (3) && eqsuiv (4))))
                         return (1);
                     else
                         return (0);
    }
}

/* compare les lignes depl1 + delta et depl2 + delta */
int eqsuiv (delta)
octet delta;
{
    int16 depl1, depl2;
    int result;


    /* sauvegarde des dplacements initiaux */
    depl1 = fichier [0].depl;
    depl2 = fichier [1].depl;

    /* comparaison des lignes */
    result = eqlig (depl1 + delta, depl2 + delta);

    /* restauration des dplacements initiaux */
    fichier [0].depl = depl1;
    fichier [1].depl = depl2;

    return (result == 1);
}

/* lecture d'une ligne */
lecture ()
{
    octet far *ligne;
    int16 szligne, szmaxlig;
    int16 deplacement;
    int   car;      /* dernier caractre lu */


    /* initialisation */
    if (ficur->depl == 0)
        deplacement = 0;
    else
        deplacement = ficur->deblig [ficur->depl - 1]
                    + ficur->szlig  [ficur->depl - 1];

    /* longueur max de la ligne en fontion de la mmoire disponible */
    if (szbuf - deplacement > SZMAXLIG)
        szmaxlig = SZMAXLIG;
    else
        szmaxlig = szbuf - deplacement;

    /* termin si en limite du buffer */
    if (szmaxlig < 5)
    {
        puts ("Dbordement de buffer : resynchronisation impossible");
        fichier [0].depl = fichier [0].nbmemo;
        fichier [1].depl = fichier [1].nbmemo;
        affdif ();
        exit ();
    }

    /* sinon, initialisation position de la ligne */
    ficur->deblig [ficur->depl] = deplacement;
    ligne = ficur->buffer + deplacement;

    /* lire le premier caractres de la ligne */
    car = getc (ficur->fic);

    /* si non fin de fichier */
    if (car == EOF)
        ficur->finfic = 1;
    else
    {
        /* on est en dbut de ligne */
        szligne = 0;

        /* tantque non fin de ligne */
        while (car != '\n' && car != EOF)
        {
            if (szligne < szmaxlig)
            {
                /* mmoriser le dernier caractre lu */
                if (car != TAB || optt == 1)
                    ligne [szligne++] = car;

                /* si pas d'option -t remplacer les tabulations par des blancs */
                else
                    do
                        ligne [szligne++] = ' ';
                    while (szligne % 8 != 0 && szligne < szmaxlig);
            }

            /* lire le caractre suivant */
            car = getc (ficur->fic);
        }

        /* supprimer le CR ventuel en fin de ligne */
        if (ligne [szligne - 1] == '\r')
            szligne --;

        /* retourner la taille de la ligne lue */
        ficur->szlig [ficur->depl] = szligne;

        /* mettre  jour le nombre de lignes mmorises */
        ficur->nbmemo ++;
    }
}

/* dition des diffrences rencontres */
affdif ()
{
    octet i;
    int   j;
    octet far *carlig;
    int16 szligne;
    char  carsens;
    int   nbaffiche [2];


    if (fichier [0].depl != 0 || fichier [1].depl != 0)
    {
        printf ("%ld,%ld\n", fichier [0].numl, fichier [1].numl);

        nbaffiche [0] = fichier [0].depl;
        nbaffiche [1] = fichier [1].depl;

        if (fichier [0].finfic == 0 && fichier [1].finfic == 0)
        {
            while (nbaffiche [0] > 1 && nbaffiche [1] > 1
                              && eqlig (nbaffiche [0] - 1, nbaffiche [1] - 1))
            {
                nbaffiche [0] = nbaffiche [0] - 1;
                nbaffiche [1] = nbaffiche [1] - 1;
            }
        }
        else
        {
            nbaffiche [0] = fichier [0].nbmemo;
            nbaffiche [1] = fichier [1].nbmemo;
        }

        for (i = 0; i < 2; i++)
        {
            ficur = &fichier [i];

            carsens = '<' + (i << 1);

            if (i == 1 && nbaffiche [0] && nbaffiche [1])
                puts ("--");

            for (j = 0; j < nbaffiche [i]; j++)
            {
                putchar (carsens);
                putchar (' ');

                carlig  = ficur->buffer + ficur->deblig [j];
                szligne = ficur->szlig [j];

                while (szligne-- > 0)
                {
                    if (*carlig != CTRZ)
                        putchar (*carlig);
                    else
                    {
                        putchar ('^');
                        putchar ('Z');
                    }

                    carlig++;
                }

                putchar ('\n');
            }

            if (fichier [1-i].finfic)
            {
                affin (i, carsens);
                fichier [i].finfic = 1;
            }
        }
    }
}

/* suppression des lignes memorises traites */
supligmemo ()
{
    octet i;
    int16 j;
    int16 k;
    int16 adrdeb;
    int16 adrfin;


    for (i = 0; i < 2; i++)
    {
        ficur = &fichier [i];

        adrdeb = ficur->deblig [ficur->depl]
               + ficur->szlig  [ficur->depl];

        adrfin = ficur->deblig [ficur->nbmemo - 1]
               + ficur->szlig  [ficur->nbmemo - 1];

        k = 0;

        for (j = ficur->depl + 1; j < ficur->nbmemo; j++)
        {
            ficur->deblig [k] = ficur->deblig [j] - adrdeb;
            ficur->szlig  [k] = ficur->szlig  [j];
            k++;
        }

        k = 0;

        for (j = adrdeb; j < adrfin; j++)
            ficur->buffer [k++] = ficur->buffer [j];

        ficur->nbmemo = ficur->nbmemo - ficur->depl - 1;
        ficur->numl   = ficur->numl   + ficur->depl + 1;
    }
}

affin (int numfic, char cardeb)
{
    int   car;      /* dernier caractre lu */


    /* initialisation */
    ficur = &fichier [numfic];

    car = getc (ficur->fic);

    while (car != EOF)
    {
        putchar (cardeb);
        putchar (' ');

        /* tantque non fin de ligne */
        while (car != '\n' && car != EOF)
        {
            /* afficher le dernier caractre lu */
            putchar (car);

            /* lire le caractre suivant */
            car = getc (ficur->fic);
        }

        /* passage  la ligne */
        putchar ('\n');

        /* lire le premier caractre de la ligne suivante */
        car = getc (ficur->fic);
    }
}
