#include <stdio.h>
#include <dir.h>
#include <alloc.h>
#include <mem.h>

/* taille d'une description des caractristiques d'un fichier */
#define szmemol    6           /* en long entiers */
#define szmemo    24           /* en octets */
#define sznfic    13           /* taille (en octets) d'un nom de fichier */
#define posmotattr 3           /* position caractristiques fichier dans table */
#define postaille  4
#define posdate    5
#define posattrib sznfic + 1   /* position en octets */


/* description des caractristiques d'un fichier */
typedef unsigned long descfic [szmemol];

static struct ffblk fich;      /* table de description des fichiers */
char droitclair [] = "adlshp"; /* droits possibles pour un fichier */

/* pointeurs sur les lments de la table de description des fichiers */
static char *attrib, *nomfich;
static unsigned int *datheure;
static long *taille;

static int optl = 0;           /* option -l de la commande diffdir */
static int trouve;             /* liste des fichiers entirement consulte ? */
static int szliste;            /* nombre maximun de fichiers mmorisables */
unsigned descfic *listefic;    /* pointeur sur liste des fichiers mmoriss */
static int nbfic = 0;          /* nombre de fichiers mmoriss */
static int debliste;           /* spare les 2 listes de fichiers */
static char nfichier [sznfic + 1]; /* nom du fichier  diter + attribut */
static unsigned long datetime; /* date et heure en format interne */

/* date et heure de mis  jour */
static unsigned char jour, mois, an, heure, min, sec;

/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    /* chemins d'accs aux fichiers du rpertoire slectionn */
    char chemin [256];

    /* dernier caractre d'un nom pass en paramtre */
    char *dercar;

    int i;  /* compteur */


    /* initialisation du nombre maximal d'lments de la liste */
    szliste = (coreleft () - 100) / szmemo;

    /* allocation de la liste des fichiers mmoriss */
    listefic = malloc (szliste * szmemo);

    /* initialisation des pointeurs sur la table de description des fichiers */
    attrib   = &fich.ff_attrib;
    nomfich  = fich.ff_name;
    datheure = &fich.ff_ftime;
    taille   = &fich.ff_fsize;

    /* calcul du nombre de paramtres de la commande */
    narg--;

    /* prise en compte de l'option -l ou -l~ ventuelle */
    if (strcmp (varg [1], "-l") == 0)
    {
        optl = 1;
        narg --;
        varg ++;
    }
    else if (strcmp (varg [1], "-l~") == 0)
    {
        optl = 2;
        narg --;
        varg ++;
    }

    /* si des noms de fichier ou de rpertoire ont t passs en paramtre */
    if (narg == 2)
    {
        /* tant que tous les paramtres n'ont pas t traits */
        for (i = 1; i <= 2; i++)
        {
            /* on considre qu'il s'agit d'un rpertoire
               fabriquer le chemin d'accs aux lments de ce rpertoire */
            dercar = varg [i] + strlen (varg [i]) - 1;

            if (*dercar == '/' || *dercar == '\\')
                sprintf (chemin, "%s*.*", varg [i]);
            else
                sprintf (chemin, "%s/*.*", varg [i]);

            /* recherche du premier fichier de ce rpertoire
               on exclue un ventuel fichier label de la recherche */
            trouve = findfirst (chemin, &fich, 0xf7);

            /* si fichier trouv */
            if (trouve != -1)
            {
                if (i == 1)
                    debliste = 0;
                else
                    debliste = nbfic;

                /* traiter l'ensemble des fichiers du rpertoire */
                lisdir ();
            }
            else
            {
                fprintf (stderr, "%s : n'est pas un rpertoire\n", varg [i]);
                exit ();
            }
        }

        /* afficher les diffrences entre les 2 rpertoires */
        affdiff ();
    }
    else
        fputs ("Syntaxe : diffdir [-l[~]] <rpertoire 1> <rpertoire 2>\n", stderr);
}


/* mmorisation de l'ensemble des fichiers du rpertoire slectionn */
lisdir ()
{
    /* si le fichier est "." */
    if (strcmp (fich.ff_name, ".") == 0)
    {
        /* passer au fichier suivant */
        trouve = findnext (&fich);

        /* si ce fichier est ".." */
        if (strcmp (fich.ff_name, "..") == 0)
            /* passer au fichier suivant */
            trouve = findnext (&fich);
    }

    /* rpter */
    do
    {
        /* mmoriser les caractristiques du fichier courant */
        memo ();

        /* chercher le fichier suivant */
        trouve = findnext (&fich);
    }
    /* jusque plus de fichier trouv */
    while (trouve != -1);
}

memo ()
{
    int i;            /* position dans la liste des fichiers */
    int ordre;        /* rsultat d'une comparaison entre 2 noms de fichier */
    descfic curfic;   /* nom du fichier courant */


    /* traitement de l'attribut du fichier */
    trtattrib ();

    /* initialisation (de 2 octets non utilises */
    curfic [posmotattr] = 0;

    /* recopier dans curfic le nom du fichier  mmoriser */
    memcpy (curfic, nomfich, sznfic);

    /* recopier l'attribut */
    memcpy ((char *) curfic + posattrib, attrib, 1);

    /* recopier l'heure et la date de mise  jour du fichier */
    memcpy (&curfic [posdate], datheure, 4);

    /* recopier la taille du fichier */
    curfic [postaille] = *taille;

    /* si la liste des fichiers est pleine */
    if (nbfic == szliste)
    {
        /* erreur fatale */
        fprintf (stderr, "Trop de fichier, commande interrompue\n");
        exit ();
    }

    /* initialisation */
    i = debliste;
    ordre = 1;

    /* chercher avant quel fichier de la liste
       le fichier courant doit tre insr */
    while ((i < nbfic) && (ordre == 1))
        ordre = compar (curfic, listefic [i++]);

    /* si on n'est pas arriv en fin de liste */
    if (ordre == -1)
    {
         /* se positionner sur le premier fichier  dplacer */
         i--;

         /* dcaler les derniers fichiers de la liste d'une position */
         memmove (listefic [i + 1], listefic [i], (nbfic - i) * szmemo);
    }

    /* mmoriser le fichier courant dans la liste */
    memcpy (listefic [i], curfic, szmemo);

    /* mise  jour du nombre de fichiers de la liste */
    nbfic++;
}


/* traitement de l'attribut du fichier courant */
trtattrib ()
{
    char i;     /* position dans le nom de fichier */


    /* se positionner a la fin du nom de fichier */
    for (i = 1; (i < sznfic) && (nomfich [i] != 0); i++);

    /* traitement de l'attribut rpertoire */
    if (*attrib & 0x10)
        nomfich [i++] = '/';

    /* on complte le nom du fichier  13 caractres */
    while (i < sznfic)
        nomfich [i++] = ' ';
}

/* comparaison de 2 noms de fichier */
int compar (x, y)
unsigned char *x, *y;
{
    int ordre = 0;     /* rsultat de la comparaison de 2 caractres */


    /* comparaison des caractres constituant les noms de fichier */
    ordre = memcmp (x, y, sznfic);

    /* retourner le rsultat de la comparaison */
    if (ordre > 0)
        return (1);

    else if (ordre < 0)
        return (-1);
    else
        return (0);
}

/* affiche les diffrences entre les 2 rpertoires */
affdiff ()
{
    int i, j;     /* position dans la liste */


    /* initialisation */
    nfichier [sznfic] = 0;
    i = 0;
    j = debliste;

    /* comparer les 2 listes de fichiers */
    while (i < debliste && j < nbfic)
    {
        switch (compar (listefic [i], listefic [j]))
        {
            case -1 : affiche (i++);
                      break;

            case  0 : if (optl)
                      {
                          if (diffcaract (listefic [i], listefic [j]))
                          {
                              affiche (i);
                              affiche (j);
                          }
                      }

                      i++;
                      j++;
                      break;

            case  1 : affiche (j++);
                      break;
        }
    }

    /* affichage de la fin de liste non explore en entier */
    while (i < debliste)
        affiche (i++);

    while (j < nbfic)
        affiche (j++);
}

/* compare les caractristiques de 2 fichiers */
int diffcaract (desc1, desc2)
char *desc1, *desc2;
{
    if (optl == 1)
        return (memcmp (desc1, desc2, szmemo));
    else
        return ((memcmp (desc1, desc2, szmemo - 4) != 0)
             || (desc1 [szmemo -2] != desc2 [szmemo - 2])
             || (desc1 [szmemo -1] != desc2 [szmemo - 1])
             || (abs (desc1 [szmemo -3] - desc2 [szmemo - 3])) > 0x10);
}

/* dition des caractristiques du (n + 1)ime fichier de la liste */
affiche (n)
int n;
{
    int masque;
    int i;


    /* prciser de quel cot se trouve le fichier qui va tre dit */
    if (n < debliste)
        putchar ('<');
    else
        putchar ('>');

    /* rcuprer et diter le nom de fichier */
    memcpy (nfichier, &listefic [n], sznfic);
    printf (" %s", nfichier);

    /* si option -l ou -l~ */
    if (optl)
    {
        /* rcuprer la date et l'heure en format interme */
        datetime = listefic [n][posdate];

        /* extraire la date de dernire mise  jour du fichier */
        jour = (datetime >> 16) & 0x1F;
        mois = (datetime >> 21) & 0x0F;
        an   = (datetime >> 25) + 80;

        /* mise  jour de l'anne  partir de l'an 2000 */
        if (an > 99)
            an -= 100;

        /* extraire l'heure de dernire mise  jour du fichier */
        heure = (datetime >> 11) & 0x1F;
        min  = (datetime >> 5) & 0x3F;
        sec  = (datetime << 1) & 0x3F;

        /* rcuprer l'attribut */
        memcpy (attrib, (char *) listefic [n] + posattrib, 1);

        /* diter les caractristiques du fichier */
        printf ("%9lu    %02d/%02d/%02d  %02d:%02d:%02d    ",
                 listefic [n][postaille], jour, mois, an, heure, min, sec);

        /* dtail de l'attribut compris */
        i = 0;

        for (masque = 0x20; masque != 0; masque = masque >> 1)
        {
            if (*attrib & masque)
                putchar (droitclair [i]);
            else
                putchar ('-');

            i++;
        }
    }

    putchar ('\n');
}
