#include <stdio.h>
#include <dos.h>

#define nbpistes     80

static struct dfree place;  /* description des caractristiques du floppy */
static char buffer [32768]; /* buffer de transfert */
static unsigned char trterr = 0; /* arrt ou message si erreur de lecture */
static unsigned char nflop; /* numro associ au lecteur de disquettes choisi */
static unsigned nbsect;     /* nombre de secteurs sur la disquette */
static unsigned secpiste;   /* nombre de secteurs par piste */
static unsigned szsect;     /* taille des secteurs sur la disquette */
static unsigned posit;      /* positionnement sur la disquette */
static unsigned pas;        /* nombre de secteurs  lire ou  recopier */
static FILE *flop;          /* descripteur de fichier image sur disque */
static char fictrav [80];   /* nom du fichier de travail */
static int  retour;         /* code de retour fonction de lecture sur disquette */


/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    char option = 0;


    /* prise en compte du nom de fichier image (en entre ou en sortie) */
    if (--narg > 0 && *varg [1] == '-')
    {
        option = varg [1][1] | ' ';

        if (option == 'r' || option == 'w' && narg > 1)
        {
            strcpy (fictrav, varg [2]);

            varg = varg + 2;
            narg = narg - 2;
        }
        else
            psyntaxe ();
    }

    /* prise en compte du nom de la disquette s'il est pass en paramtre */
    if (narg > 0)
    {
        if (narg == 1 && (varg [1][1] == 0 || varg [1][1] == ':'))
            nflop = (*varg [1] & 0x1f) - 1;
        else
            psyntaxe ();
    }
    else
        /* par dfaut, la disquette est A: */
        nflop = 0;

    /* si utilisation d'un fichier image */
    if (option)
    {
        /* effectuer l'opration demande */
        if (option == 'w')
            lecture ();
        else
        {
            printf ("Mettre la disquette de destination et taper \"return\" ");
            anadsk ();
            ecriture ();
        }
    }
    /* sinon, duplication de disquettes */
    else
    {
        /* rcupration ou cration du rpertoire des fichiers de travail */
        strcpy (fictrav, getenv ("TEMP"));

        if (*fictrav == 0)
            strcpy (fictrav, "c:/tmp");

        /* cration du rpertoire si ncessaire */
        mkdir (fictrav);

        /* fabrication du chemin d'accs au fichier de travail */
        strcat (fictrav, "/floppy");

        /* boucle de travail */
        do
        {
            lecture ();
            ecriture ();

            printf ("Voulez vous copier une autre disquette ? ");
        }
        /* jusque plus de copie */
        while (lirep () != 'n');

        /* dtruire le fichier image */
        fin ();
    }
}



/* rappel de la syntaxe de la commande */
psyntaxe ()
{
    fprintf (stderr,
        "Syntaxe : dupliflo [-(r|w) <nom fichier image>] [<nom de disque>]\n");

    exit ();
}



/* lecture disquette  copier et fabrication du fichier image */
lecture ()
{
    printf ("Mettre la disquette  copier et taper \"return\" ");

    /* analyser les caractristiques de la disquette */
    anadsk ();

    /* cration et ouverture du fichier image */
    flop = fopen (fictrav, "wb");

    /* traitement d'erreur fatale */
    if (flop == 0)
    {
        fprintf (stderr, "Impossible de crer le fichier image %s\n", fictrav);
        exit ();
    }

    /* initialisation */
    pas   = 64;

    /* tantque la disquette n'a pas t lue entirement */
    while (posit < nbsect)
    {
        /* calculer la taille de la zone  lire */
        if (posit + pas > nbsect)
            pas = nbsect - posit;

        /* lire une partie de la disquette */
        retour = absread (nflop, pas, posit, buffer);

        /* si erreur de lecture */
        if (retour < 0)
        {
            if (trterr == 0)
            {
                puts ("Erreur de lecture");
                printf ("Voulez vous continuer malgr les erreurs ? ");

                /* prise en compte de la rponse */
                if (lirep () == 'n')
                    fin ();
                else
                    trterr = 1;
            }

            pas = 1;

            memset (buffer, 0x7F ,szsect);

            retour = absread (nflop, pas, posit, buffer);

            if (retour < 0)
                printf ("Erreur de lecture aprs %d Koctets\n",
                                 (long) posit * szsect / 1024);
        }

        /* la recopier sur disque */
        if (fwrite (buffer, szsect, pas, flop) < pas)
        {
            /* traitement d'erreur */
            perror ("fwrite");
            fin ();
        }

        /* recalculer la position de la zone non lue */
        posit = posit + pas;

        /* augmenter le pas, si rduit sur des erreurs de lecture */
        if (retour >= 0 && pas <= 32)
            pas = pas + pas;
    }

    /* fermer le fichier image */
    fclose (flop);
}



/* analyse des caractristiques d'une disquette */
anadsk ()
{
    /* attente d'un return */
    while (getchar () != '\n');

    /* lecture des caractristiques de la disquette */
    getdfree (nflop + 1, &place);

    /* vrification avant lecture */
    if (place.df_sclus < 0)
    {
        fprintf (stderr, "Disquette manquante ou non formatte\n");
        exit ();
    }

    /* calcul du nombre de secteurs par pistes */
    secpiste = place.df_total * place.df_sclus / nbpistes;

    if ((place.df_total * place.df_sclus) % nbpistes != 0)
        secpiste++;

    /* calcul du nombre total de secteur et rcupration de leur taille */
    nbsect = secpiste * nbpistes;
    szsect = place.df_bsec;

    /* initialisation */
    posit =  0;
}



/* lecture du fichier image et fabrication des nouvelles disquette */
ecriture ()
{
    /* ouvrir le fichier image en lecture */
    flop = fopen (fictrav, "rb");

    /* traitement d'erreur fatale */
    if (flop == 0)
    {
        fprintf (stderr, "Impossible de lire le fichier image %s\n", fictrav);
        exit ();
    }

    do
    {
        /* si on a dj fait une disquette */
        if (posit != 0)
        {
            /* se repositionner au dbut du fichier image */
            rewind (flop);
            posit = 0;

            /* attendre que l'oprateur ait chang de disquette */
            printf ("Changer la disquette et taper \"return\" ");
            while (getchar () != '\n');
        }

        /* initialisation */
        pas = 64;

        /* tantque la copie n'est pas finie */
        while (pas == 64)
        {
            /* lire un morceau du fichier image */
            pas = fread (buffer, szsect, pas, flop);

            /* le recopier sur la disquette */
            if (abswrite (nflop, pas, posit, buffer) < 0)
            {
                if (posit == 0)
                {
                    fputs ("Disquette manquante ou non formatte\n", stderr);
                    posit ++;
                }
                else
                    fputs ("Capacit de la disquette trop faible\n", stderr);

                pas = 0;
            }

            /* recalculer la position de la zone non lue */
            posit = posit + pas;
        }

        /* nouvelle copie ? */
        printf ("Voulez vous faire une autre copie de la mme disquette ? ");
    }
    /* jusque plus de copie */
    while (lirep () != 'n');

    /* fermer le fichier image */
    fclose (flop);
}



/* lecture de la rponse  une question */
int lirep ()
{
    char reponse;


    /* lecture de la rponse */
    reponse = getchar ();

    /* vider le buffer clavier jusqu'au "return" */
    if (reponse != '\n')
        while (getchar () != '\n');

    /* retourner la rponse */
    return (reponse | 0x20);
}



/* fermeture et destruction du fichier image */
fin ()
{
    fclose (flop);
    unlink (fictrav);

    /* sortir du programme */
    exit ();
}
