#include <stdio.h>
#include <dir.h>
#include <dos.h>
#include <io.h>


#define octet unsigned char
#define int32 unsigned long

#define infini 0xFFFFFFFF

#define sznfich    13
#define szlistini 128

/* dfinition type nom de fichier */
typedef char nomfic [sznfich + 1];

/* nombre de noms de fichier sur lesquels s'effectue de slection */
static octet nbname  = 0;

/* noms de fichiers sur lesquels s'effectue la slection */
static char *nfich [16];

/* taille min et max des fichiers  slectionner */
static int32 minsize = 0;
static int32 maxsize = infini;

/* age min et (max en secondes) des fichiers  slectionner */
static int32 agemin  = 0;
static int32 agemax  = infini;

/* date de mise  jour min et max des fichiers  slectionner */
static int32 minmtime = 0;
static int32 maxmtime = infini;


/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    octet nbacces = 0;  /* nombre de rpertoires de dmarrage */


    /* test du nombre d'arguments */
    if (--narg < 1)
        psyntaxe ();

    /* se positionner sur le premier argument */
    varg++;

    /* compter le nombre de rpertoires de dmarrage */
    do
        nbacces++;
    while (*varg [nbacces] != '-' && nbacces < narg);

    /* prise en compte des options de la commande */
    if (narg > nbacces)
        trtopt (narg - nbacces, varg + nbacces);

    /* lancer la recherche sur chacun des rpertoires de dmarrage */
    tsexplore (nbacces, varg);
}

/* traitement des options de la commande */
trtopt (narg, varg)
int narg;
char *varg [];
{
    char  signe;    /* indique si la valeur est une limite inf ou sup */
    int32 valeur;   /* taille ou anciennet du fichier */
    char  unite;    /* unite dans laquelle est exprime valeur */
    int32 valunit;  /* coefficient multiplicateur de l'unit de mesure */
    char *carname;  /* caractre du nom de fichier  trouver */


    /* tantqu'il y a des arguments  traiter */
    do
    {
        /* option nom de fichier ? */
        if (strcmp (*varg, "-name") == 0)
        {
            if (narg == 1)
            {
                fprintf (stderr, "Nom de fichier manquant\n");
                exit ();
            }

            /* rcuprer la liste de noms de fichiers */
            do
            {
                varg++;

                /* mmoriser l'adresse du nom de fichier */
                carname = *varg;
                nfich [nbname++] = carname;

                /* mettre ce nom de fichier en majuscules */
                do
                {
                    if (*carname >= 'a' && *carname <= 'z')
                        *carname = *carname - ' ';

                    /* passer au caractre suivant */
                    carname++;
                }
                while (*carname != 0);
            }
            while (--narg > 1 && *varg [1] != '-');
        }

        /* option anciennet du fichier ? */
        else if (strcmp (*varg, "-mtime") == 0)
        {
            if (--narg == 0)
            {
                fprintf (stderr, "Anciennet du fichier manquante\n");
                exit ();
            }

            /* rcuprer la valeur de l'option anciennet du fichier */
            getval (*(++varg), &signe, &valeur, &unite);

            /* prise en compte de l'unit */
            switch (unite)
            {
                case ' ' :
                case 'j' : valunit = 86400;
                           break;

                case 'h' : valunit = 3600;
                           break;

                case 'm' : valunit = 60;
                           break;

                case 's' : valunit = 1;
                           break;

                default  : signe = 0;  /* pour genrer une erreur */
            }

            /* mettre  jour les limites d'anciennet des fichiers */
            majlimit (&agemin, &agemax, signe, valeur, valunit, varg);
        }

        /* option taille du fichier */
        else if (strcmp (*varg, "-size") == 0)
        {
            if (--narg == 0)
            {
                fprintf (stderr, "Taille du fichier manquante\n");
                exit ();
            }

            /* rcuprer la valeur de l'option taille du fichier */
            getval (*(++varg), &signe, &valeur, &unite);

            /* prise en compte de l'unit */
            switch (unite)
            {
                case 'k' : valunit = 1024;
                           break;

                case ' ' : valunit = 1;
                           break;

                default  : signe = 0;  /* pour generer une erreur */
            }

            /* mettre  jour les limites de taille des fichiers */
            majlimit (&minsize, &maxsize, signe, valeur, valunit, varg);
        }

        /* option inconnue */
        else
        {
            fprintf (stderr, "Option %s incorrecte\n", *varg);
            psyntaxe ();
        }

        /* se positionner sur l'argument suivant */
        varg++;
    }
    while (--narg > 0);

    /* si option anciennet du fichier slectionne */
    if (agemin != 0 || agemax != infini)
        /* mettre en forme les date limites */
        majmtime ();
}

/* rcupration de la taille ou l'anciennet des fichiers  slectionner */
getval (chaine, signe, valeur, unite)
char  *chaine;
char  *signe;
int32 *valeur;
char  *unite;
{
    /* rcuprer le signe */
    if (*chaine == '+' || *chaine == '-')
        *signe = *(chaine ++);
    else
        *signe = ' ';

    /* rcuprer la valeur */
    *valeur = 0;

    while ('0' <= *chaine && *chaine <= '9')
        *valeur = (*valeur * 10) + *(chaine++) - '0';

    /* rcuprer l'unit */
    *unite = *chaine | ' ';
}

/* mise  jour des limites de taille ou d'anciennet des fichiers */
majlimit (min, max, signe, valeur, valunit, options)
int32 *min, *max;
char signe;
int32 valeur, valunit;
char **options;
{
    switch (signe)
    {
        case ' ' : if (*min == 0 && *max == infini)
                   {
                       *min = valeur * valunit;
                       *max = (valeur + 1) * valunit - 1;
                   }
                   else
                   {
                       fprintf (stderr,
                  "Valeurs concurrentes pour l'option %s\n", *(options - 1));

                       exit ();
                   }
                   break;

        case '+' : if (*min == 0)
                       *min = (valeur + 1) * valunit;
                   else
                   {
                       fprintf (stderr,
                  "Valeurs concurrentes pour l'option %s\n", *(options - 1));

                       exit ();
                   }
                   break;

        case '-' : if (*max == infini)
                       *max = (valeur * valunit) - 1;
                   else
                   {
                       fprintf (stderr,
                  "Valeurs concurrentes pour l'option %s\n", *(options - 1));

                       exit ();
                   }
                   break;

        default  : fprintf (stderr, "Valeur %s incorrecte pour l'option %s\n",
                                                   *options, *(options - 1));
                   psyntaxe ();
    }
}

/* mise en forme des dates limites de mise  jour des fichiers */
majmtime ()
{
    int32 maintenant;  /* nombre de secondes depuis le 1 janv 1970 */


    /* rcuprer la date et l'heure courante au format unix */
    time (&maintenant);

    /* mise  jour si ncessaire de l'anciennet minimale */
    if (agemin != 0)
        convmtime (&maxmtime, maintenant - agemin);

    /* mise  jour si ncessaire de l'anciennet maximale */
    if (agemax != infini)
        convmtime (&minmtime, maintenant - agemax);
}


/* conversion d'une valeur limite d'anciennet des fichiers */
convmtime (timefic, valtime)
struct ftime *timefic;
int32 valtime;
{
    struct date dosdate;    /* date au format dos */
    struct time dostime;    /* heure au format dos */


    /* convertir valtime au format ms-dos */
    unixtodos (valtime, &dosdate, &dostime);

    /* mmoriser la date et l'heure limite */
    timefic->ft_year  = dosdate.da_year - 1980;
    timefic->ft_month = dosdate.da_mon;
    timefic->ft_day   = dosdate.da_day;
    timefic->ft_hour  = dostime.ti_hour;
    timefic->ft_min   = dostime.ti_min;
    timefic->ft_tsec  = dostime.ti_sec / 2;
}

/* test d'existence des rpertoires et lancement de la recherche */
tsexplore (nbacces, varg)
int nbacces;
char *varg [];
{
    char *cardir;       /* dernier caractre du chemin d'accs */
    char chemin [80];   /* chemin d'accs aux fichiers de ce rpertoire */
    struct ffblk fich;  /* structure d'un lment de rpertoire */


    /* pour chacun des chemins d'accs */
    while (nbacces-- > 0)
    {
        cardir = *varg;

        /* normaliser le chemin d'accs introduit */
        do
        {
            /* passer les lettres en majuscules */
            if ('a' <= *cardir && *cardir <= 'z')
                *cardir = *cardir - ' ';

            /* le '/' est le symbole de rpertoire */
            else if (*cardir == '\\')
                *cardir = '/';

            /* caractre suivant */
            cardir++;
        }
        while (*cardir != 0);

        /* supprimer le dernier '/' ventuel */
        cardir--;
        if (*cardir == '/')
            *cardir = 0;

        /* si on ne cherche pas  partir de la racine */
        if (**varg != 0)
        {
            /* rcuprer le rpertoire de dmarrage de la recherche */
            sprintf (chemin, "%s/*.*", *varg);

            /* si c'est un rpertoire, l'explorer */
            if (findfirst (chemin, &fich, 0xFF) == 0)
                explore (*varg);

            /* sinon message d'erreur */
            else
                fprintf (stderr, "%s n'est pas un rpertoire\n", *varg);
        }
        /* sinon aucun test de validit n'est ncessaire */
        else
            explore ("");

        /* passer  l'argument suivant */
        varg++;
    }
}

/* comparaison du nom de fichier avec un nom de la liste
   ou test de la fin du nom du fichier */
octet cmpnfich (carfic, carlis)
char *carfic, *carlis;
{
    /* comparer les 2 noms caractres par caractres */
    do
    {
        /* si 2 caractres identiques ou mtacaractre '?' accept */
        if (*carfic == *carlis || (*carlis == '?' && *carfic != 0))
        {
            /* controler les 2 caractres suivants */
            carfic++;
            carlis++;
        }
        /* sinon si mtacaractre '*' */
        else if (*carlis == '*')
        {
            /* se positionner apres le(s) caractre(s) '*' */
            do
                carlis++;
            while (*carlis == '*');

            /* terminsi aucun caractre apr '*' */
            if (*carlis == 0)
                return (1);

            /* sinon */
            else
            {
                /* test de la fin du nom de fichier */
                while (*carfic != 0)
                    if (cmpnfich (carfic, carlis) == 1)
                        return (1);
                    else
                        carfic++;

                /* echec de la comparaison */
                return (0);
            }
        }
        /* sinon echec dans la comparaison */
        else
            return (0);
    }
    while (*carlis != 0);

    /* analyse termine, resultat en fonction longueur du nom de fichier */
    if (*carfic == 0)
        return (1);
    else
        return (0);
}


/* controle de validit d'un nom de fichier */
octet tsname (fichier)
char fichier [];
{
    octet poslis;  /* position dans la liste des noms de fichiers */


    /* comparer le nom du fichier avec ceux de la liste */
    for (poslis = 0; poslis < nbname; poslis++)
        if (cmpnfich (fichier, nfich [poslis]) == 1)
            return (1);

    /* termine si le nom du fichier n'est pas dans la liste autorise */
    return (0);
}

/* exploration du rpertoire chaine et de son arborescence */
explore (chaine)
char chaine [];
{
    char chemin [256];   /* chemin d'accs aux fichiers */
    struct ffblk fich;   /* structure d'un lment de rpertoire */
    octet trouve;        /* indique si fichier trouv dans l'arborescence */
    octet select;        /* indique si le fichier trouv est slectionn */
    int32 anciennete;    /* fusion fich.ff_fdate et fich.ff_ftime */
    nomfic *listefic, *newliste; /* liste de noms de fichiers slectionns */
    int szlistefic;      /* tespace mmoire allou pour la liste de fichiers */
    int nbfic, posliste; /* nombre de fichiers et position dans la liste */


    /* fabriquer le chemin d'accs aux fichiers du rpertoire  explorer */
    sprintf (chemin, "%s/*.*", chaine);

    /* exploration du rpertoire */
    trouve = findfirst (chemin, &fich, 0xFF);

    /* on saute les fichiers . et .. */
    if (strcmp (fich.ff_name, ".") == 0)
    {
        trouve = findnext (&fich);

        if (strcmp (fich.ff_name, "..") == 0)
            trouve = findnext (&fich);
    }

    /* initialisation : aucun fichier trouv */
    nbfic = 0;
    szlistefic = szlistini;

    /* allocation de listefic */
    listefic = (nomfic *) calloc (szlistefic, sznfich + 1);

    if (listefic == 0)
    {
        fprintf (stderr, "Dbordement mmoire\n");
        exit ();
    }

    /* tantque rpertoire non explor en entier */
    while (trouve == 0)
    {
        /* mmorisation anciennet du fichier sous forme d'un entier long */
        memcpy (&anciennete, &fich.ff_ftime, 4);

        /* test de validit de la taille, de la date de mise  jour
                                               et du nom de fichier */
        if (minsize <= fich.ff_fsize && fich.ff_fsize <= maxsize
             && minmtime <= anciennete && anciennete <= maxmtime
                 && (nbname == 0 || tsname (fich.ff_name) == 1))
            select = 1;
        else
            select = 0;

        /* on slectionne aussi les rpertoires pour continuer l'exploration */
        select = select | (fich.ff_attrib & 0x10);

        /* on ne slectionne pas les fichiers spciaux Windows 95 */
        if (fich.ff_attrib == 0x0F)
            select = 0;

        /* mmorisation des fichier slectionns dans l'ordre alphabtique */
        if (select != 0)
        {
            /* si liste de fichiers trop petite */
            if (nbfic >= szlistefic)
            {
                newliste = (nomfic *) realloc (listefic, (szlistefic + szlistini) * (sznfich + 1));

                /* prise en compte du rsultat de l'opration */
                if (newliste != 0)
                {
                    /* nouvelle taille et nouvelle adresse de la liste */
                    szlistefic = szlistefic + szlistini;
                    listefic = newliste;
                }
                else
                {
                    /* dbordement, on traite le premier fichier class */
                    select = listefic [0][sznfich];

                    /* edition des fichiers slectionns */
                    if (select & 1)
                        printf ("%s/%s\n", chaine, listefic [0]);

                    /* exploration des sous rpertoires */
                    if (select & 0x10)
                    {
                        sprintf (chemin, "%s/%s", chaine, listefic [0]);

                        explore (chemin);
                    }

                    /* et on le supprime de la liste */
                    memmove (listefic [0], listefic [1], --nbfic * sznfich);
                }
            }

            /* recherche de l'endroit o insrer le nom de fichier */
            for (posliste = 0; (posliste < nbfic) &&
                (strcmp (listefic [posliste], fich.ff_name)) < 0; posliste++);

            if (posliste < nbfic)
                memmove (listefic [posliste + 1], listefic [posliste],
                                            (nbfic - posliste) * (sznfich + 1));

            /* mmorisation du nom de fichier */
            memmove (listefic [posliste], fich.ff_name, sznfich);

            /* mmorisation de l'attribut de slection */
            listefic [posliste][sznfich] = select;

            /* un fichier de plus est mmoris */
            nbfic++;
        }

        /* passer  l'lment de rpertoire suivant */
        trouve = findnext (&fich);
    }

    /* dition des fichier slectionns et exploration des sous rpertoires */
    for (posliste = 0; posliste < nbfic; posliste++)
    {
        select = listefic [posliste][sznfich];

        /* edition des fichiers slectionns */
        if (select & 1)
            printf ("%s/%s\n", chaine, listefic [posliste]);

        /* exploration des sous rpertoires */
        if (select & 0x10)
        {
            sprintf (chemin, "%s/%s", chaine, listefic [posliste]);

            explore (chemin);
        }
    }

    /* libration mmoire */
    free (listefic);
}


/* rappel de la syntaxe de la commande */
psyntaxe ()
{
    fprintf (stderr, "%s\n%s\n%s\n",
    "Syntaxe : find <nom(s) de rpertoire(s)> [-name <nom de fichier>]",
    "                                [-size [+|-]<taille du fichier>[k]]",
    "                     [-mtime [+|-]<anciennet du fichier>[j|h|m|s]]");

    exit ();
}
