#include <stdio.h>
#include <io.h>
#include <dir.h>
#include <alloc.h>

#define szficdef     14
#define szlistini    64

#define octet char


/* nom d'un fichier + nature */
typedef char ficdef [szficdef];

/* date et heure au lancement de la commande */
static unsigned long hdebut;

/* date de dernire mise  jour du fichier  dplacer */
static unsigned long mtime;

/* chemin d'accs aux fichiers du rpertoire slectionn */
char chemin [256];


/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    /* table de description d'un fichier */
    struct ffblk fich;


    /* rcuprer la date et l'heure courantes */
    recuptime ();

    /* test du nombre d'arguments */
    if (narg != 3)
        erreur ("Syntaxe : mvdir <rpertoire d'origine>",
                                           " <rpertoire de destination>", 0);

    /* mettre en forme les arguments */
    mise_en_forme (varg [1]);
    mise_en_forme (varg [2]);

    /* pas de dplacement entre disques diffrents */
    if ((varg [1][1] == ':') || (varg [2][1] == ':'))
        if ((varg [1][0] != varg [2][0]) || (varg [1][1] != varg [2][1]))
            erreur ("Les noms de disque doivent tre identiques", 0, 0);

    /* vrifier que les 2 arguments sont des rpertoires */
    if (access (varg [1], 0) != 0)
        erreur ("Rpertoire ", varg [1], " inexistant");

    if ((findfirst (varg [1], &fich, 0xff) == 0)
               && ((fich.ff_attrib & 0x10) == 0))
        erreur (varg [1], " est un fichier", 0);

    /* si le rpertoire de destination n'existe pas encore */
    if (access (varg [2], 0) != 0)
    {
        /* essayer un dplacement ultrarapide */
        if (rename (varg [1], varg [2]) == 0)
            exit ();

        if (mkdir (varg [2]) != 0)
            erreur ("Impossible de crer le rpertoire ", varg [2], 0);
    }
    /* sinon, le rpertoire de destination existe dja */
    else
    {
        if ((findfirst (varg [2], &fich, 0xff) == 0)
                     && ((fich.ff_attrib & 0x10) == 0))
            erreur (varg [2], " existe dj comme fichier", 0);

        if (strcmp (varg [1], varg [2]) == 0)
            erreur ("L'origine et la destination sont identiques", 0, 0);

        if ((strlen (varg [1]) < strlen (varg [2])) ||
            (memcmp (varg [1], varg [2], strlen (varg [2])) != 0))
            erreur ("Le rpertoire ", varg [2], " existe dj");
    }

    /* dplacement des fichiers */
    mvdir (varg [1], varg [2]);
}


/* message d'erreur fatal */
erreur (chaine1, chaine2, chaine3)
char *chaine1, *chaine2, *chaine3;
{
    fputs (chaine1, stderr);

    if (*chaine2 != 0)
        fputs (chaine2, stderr);

    if (*chaine3 != 0)
        fputs (chaine3, stderr);

    putc ('\n', stderr);

    exit ();
}


/* mettre "chaine" sous forme normalise */
mise_en_forme (chaine)
char chaine [];
{
    while (*chaine != 0)
    {
        /* remplacer les caractres '\' par des '/' */
        if (*chaine == '\\')
            *chaine = '/';

        /* et les caractres minuscules par la majuscule correspondante */
        else if (*chaine >= 'a')
            *chaine = *chaine - 0x20;

        /* passer au caractre suivant */
        chaine++;
    }
}


/* rcupration de la date et de l'heure courante */
recuptime ()
{
    unsigned char heure [4];
    unsigned int jour [2];

    /* rcupration de la date et de l'heure */
    getdate (jour);
    gettime (heure);

    /* conversion dans le format employ pour les fichiers */
    hdebut = ((jour [0] - 1980) << 9) | ((jour [1] & 0xff00) >> 3)
                                               | (jour [1] & 0xff);
    hdebut = (hdebut << 16) | (((heure [1] << 11) | (heure [0] << 5)
                              | (heure [3] >> 1)) & 0xffff);
}


/* dplacement rpertoire "orig" et de son arborescence dans "dest" */
mvdir (orig, dest)
char orig [], dest [];
{
    /* chemin d'accs aux fichiers des diffrents rpertoires */
    char *dirorig, *dirdest;
    char *ficorig, *ficdest;

    /* table de description d'un fichier */
    struct ffblk fich;

    /* indicateur */
    int trouve;

    /* compteurs */
    int posliste, nbfic, szliste;

    /* liste des noms et attribut des fichiers d'un rpertoire */
    ficdef *listefic, *newliste;


    /* allouer les chaines avec la bonne taille mmoire */
    dirorig = malloc (strlen (orig) + 1);
    dirdest = malloc (strlen (dest) + 1);
    ficorig = malloc (strlen (orig) + szficdef);
    ficdest = malloc (strlen (dest) + szficdef);

    /* allocation du tableau listefic */
    szliste = szlistini;
    listefic = malloc (szliste * szficdef);

    /* fabriquer le chemin d'accs aux lments du rpertoire  deplacer */
    strcpy (chemin, orig);
    strcat (chemin, "/*.*");

    /* on exclue un ventuel fichier label de la recherche */
    trouve = findfirst (chemin, &fich, 0xf7);

    /* si ce fichier est "." */
    if (strcmp (fich.ff_name, ".") == 0)
    {
        /* passer au fichier suivant */
        trouve = findnext (&fich);

        /* si ce fichier est ".." */
        if (strcmp (fich.ff_name, "..") == 0)
            /* passer au fichier suivant */
            trouve = findnext (&fich);
    }

    /* initialisation : nombre de fichiers mmoriss */
    nbfic = 0;

    /* tantque on trouve des nouveaux fichiers dans le rpertoire */
    while (trouve == 0)
    {
        /* si liste de fichiers trop petite */
        if (nbfic >= szliste)
        {
            /* augmenter la taille de la liste */
            szliste  = szliste + szlistini;
            newliste = realloc (listefic, szliste * szficdef);

            /* prise en compte du rsultat de l'opration */
            if (newliste != 0)
            {
                if (listefic != newliste)
                    free (listefic);

                listefic = newliste;
            }
            else
            {
                erreur ("Dbordement mmoire", 0, 0);
                exit ();
            }
        }

        /* rcuprer la date de cration du fichier */
        memcpy (&mtime, &fich.ff_ftime, 4);

        /* si ce n'est pas un rpertoire cr pendant la copie */
        if (mtime < hdebut)
        {
            /* chercher l'lment de la liste o le nouveau fichier
               doit tre insr pour respecter l'ordre alphabtique */
            for (posliste = 0; (posliste < nbfic) &&
                (strcmp (listefic [posliste], fich.ff_name)) < 0; posliste++);

            /* si la liste n'a pas t parcourrue en entier */
            if (posliste < nbfic)
                /* dcaler les fichiers restants d'une position */
                memmove (listefic [posliste + 1], listefic [posliste],
                                              (nbfic - posliste) * szficdef);

            /* recopier le nom du fichier dans la liste */
            memmove (listefic [posliste], fich.ff_name, 13);

            /* recopier l'attribut du fichier */
            listefic [posliste][13] = fich.ff_attrib;

            /* incrmenter le nombre de fichiers mmoriss */
            nbfic++;
        }

        /* rechercher le fichier suivant du rpertoire */
        trouve = findnext (&fich);
    }

    /* construire les chemins d'accs aux rpertoires suivis d'un / */
    strcpy (dirorig, orig);

    if (dirorig [strlen (dirorig) - 1] != '/')
        strcat (dirorig, "/");

    strcpy (dirdest, dest);
    
    if (dirdest [strlen (dirdest) - 1] != '/')
        strcat (dirdest, "/");

    /* pour tous les fichiers mmoriss */
    for (posliste = 0; posliste < nbfic; posliste++)
    {
        /* construire les chemins d'accs aux fichiers */
        strcpy (ficorig, dirorig);
        strcat (ficorig, listefic [posliste]);

        strcpy (ficdest, dirdest);
        strcat (ficdest, listefic [posliste]);

        /* s'il s'agit d'un rpertoire */
        if (listefic [posliste][13] & 0x10)
        {
            /* crer le rpertoire de destination s'il n'existe pas */
            if (access (ficdest, 0) == -1)
                mkdir (ficdest);

            /* dplacer l'arborescence du rpertoire d'origine */
            mvdir (ficorig, ficdest);
        }
        /* sinon (fichier classique) */
        else
        {
            /* le dplacer */
            if (rename (ficorig, ficdest) != 0)
                erreur ("Erreur de dplacement du fichier ", ficorig, 0);
        }
    }

    /* librer l'espace mmoire allou */
    free (ficorig);
    free (ficdest);
    free (dirorig);
    free (dirdest);
    free (listefic);

    /* dtruire le rpertoire d'origine (si possible) */
    rmdir (orig);
}
