/*
 Programme sort (version terminale)

   Il est recommand de compiler ce programme avec le modle mmoire Compact
   pour avoir le maximum de mmoire afin de mmoriser le(s) fichier(s)  trier

   Deuxime amlioration :
   - allocation de presque toute la mmoire disponible au lancement du programme
   - utilisation de cette mmoire pour charger les lignes de texte
   - puis au moment du tri pour fabriquer les sous listes et compter
     leurs occurences
   - on ne conserve que les sous listes non vides avant le passage au niveau
     de tri suprieur.

   En procdant ainsi, on vite la perte de quelques octets  chaque
   allocation mmoire.
*/

#include <stdio.h>
#include <alloc.h>
#include <mem.h>

#define memolibre   1000

#define octet unsigned char

static octet optd = 0;      /* tri dictionnaire jeu de caractres MS-DOS */
static octet optw = 0;      /* tri dictionnaire jeu de caractres Windows */
static octet optn = 0;      /* option tri numrique */
static octet optr = 0;      /* option tri ordre inverse */

static int posdeb = 0;      /* colonne de dbut de tri */
static int posfin = 0x7FFB; /* colonne de fin de tri */
static int posdebl;         /* position 1er caractre dans ligne mmorise */

octet huge *memoire;        /* chaines de caractres  trier et tables de tri */
octet huge *souslistes;     /* sous listes partiellement tries */
octet huge *occurences;     /* taille de ces sous listes */

static long taillemem;      /* taille de la zone mmoire alloue */
static long positmem;       /* position dans la zone nmoire alloue */
static long tailletables;   /* longueur max zone mmoire sous listes et occur. */

/* prototypes */
octet valcar (octet caract);
psyntaxe (char *option);
tri (long posmem, long postables, int posit);
long adrliste (octet huge *adrtable, int index);
tricouple (long positlig, int posit);
octet valcar (octet caract);
edition (long posmem);
editligne (long posmem);

/* programme principal */
main (narg, varg)
int narg;
char *varg [];
{
    int valopt;  /* pour rcupration d'un argument de position */

    /* positionnement sur le premier argument */
    varg++;
    narg--;

    /* traitement des options de la commande */
    while (**varg == '-' || atoi (*varg) != 0)
    {
        valopt = atoi (*varg);

        if (valopt > 0)
            posdeb = valopt - 1;

        else if (valopt < 0)
            posfin = -valopt - 1;

        else if (varg [0][2] != 0)
            psyntaxe (*varg);
        else
        {
            switch (varg [0][1])
            {
                /* tri alphabtique, caractres accentus MS-DOS */
                case 'd' : optd = 1;
                           break;

                /* tri alphabtique, caractres accentus Windows */
                case 'w' : optw = 1;
                           break;

                /* tri numrique */
                case 'n' : optn = 1;
                           break;

                /* inversion de l'ordre de tri */
                case 'r' : optr = 1;
                           break;

                default  : psyntaxe (*varg);
            }
        }

        /* passer  l'argument suivant */
        varg++;
        narg--;
    }

    /* test de la validit des options mmorises */
    if (optn != 0 && (optd != 0 || optw != 0))
    {
        fputs ("Option -n incompatible avec les options -d et -w\n", stderr);
        exit (-1);
    }

    if (posdeb > posfin)
    {
        printf ("posdeb = %d, posfin = %d\n", posdeb, posfin);
        fputs ("Erreur dans la dfinition de la zone de tri\n", stderr);
        psyntaxe (NULL);
    }

    /* initialisation : pas encore de ligne  trier */
    posdebl = 5 + (optn * 4);

    /* rserver la quasi totalit de la mmoire */
    taillemem = farcoreleft () - memolibre;
    memoire = farmalloc (taillemem);

    /* test de la mmoire restante */
    if (memoire == 0)
    {
        /* erreur fatale */
        fputs ("Erreur fatale d'allocation mmoire\n", stderr);
        exit (-5);
    }

    /* positionnement au dbut de la zone mmoire */
    positmem = 0;

    /* pour tous les fichiers passs en paramtre */
    do
    {
        /* si au moins un fichier, y accder en lecture,
           sinon utiliser l'entre standard */
        if (narg > 0)
        {
            if (freopen (*varg, "rb", stdin) == 0)
                fprintf (stderr, "Fichier %s inexistant\n", *varg);
        }

        /* mmoriser les lignes du fichier (ou de l'entre standard) */
        chargefic ();

        /* passer au fichier suivant */
        varg++;
        narg--;
    }
    while (narg > 0);

    /* si donnes  trier */
    if (positmem > 0)
    {
        /* prparation du tri avec l'option -n */
        if (optn)
            recupnum ();

        tailletables = (taillemem - positmem) / 4;
        occurences = memoire + positmem;
        souslistes = occurences + tailletables;

        /* trier les lignes mmorises */
        tri (0, 0, posdeb);
    }

    /* pour faire propre, on libre la mmoire utilise */
    farfree ((void *) memoire);
}



/* rappel de la syntaxe de la commande */
psyntaxe (char *option)
{
    if (option)
        fprintf (stderr, "Option %s incorrecte\n", option);

    fprintf (stderr, "%s\n                               %s\n",
             "Syntaxe : sort [-r] [-(n|d|w)] [+<colonne dbut tri>]",
             "[-<colonne fin tri>] [<liste fichiers>]\n");
    exit (-1);
}



/* mmorisation des lignes d'un fichier (ou de l'entre standard) */
chargefic ()
{
    int  tailleligne;       /* taille de la ligne lue */
    int  car;               /* caractte lu */
    static long posligprec = -1;   /* numro d'octet du dbut de la ligne prcdente */
    long posentete;         /* numro d'octet du dbut de la ligne courante */

    /* lire un caractre */
    car = getchar ();

    /* tant que non fin de fichier */
    while (car != EOF)
    {
        /* initialisation */
        tailleligne = 0;
        posentete = positmem;

        /* rserver la place pour l'adresse de la
           ligne suivante et la taille de la ligne */
        positmem = positmem + posdebl;

        /* tant que non fin de ligne */
        while (car != '\n' && car != EOF)
        {
            /* mmoriser le caractre lu */
            if (tailleligne < 0x7FFF)
            {
                *(memoire + positmem) = car;
                tailleligne++;
                positmem++;

                /* test de la mmoire restante */
                if (positmem >= taillemem)
                {
                    /* erreur fatale */
                    fprintf (stderr, "Fichier trop gros, utiliser gsort\n");
                    exit (-2);
                }
            }

            /* et lire le caractre suivant */
            car = getchar ();
        }

        /* pas encore d'autre ligne aprs celle ci */
        *(memoire + posentete + 2) = -1;

        /* mmoriser la longueur de la dernire ligne lue */
        memcpy ((void *)(memoire + posentete + 3), &tailleligne, 2);

        /* rajouter cette ligne dans la liste des lignes mmorises */
        if (posligprec >= 0)
            memcpy ((void *)(memoire + posligprec), &posentete, 3);

        posligprec = posentete;

        /* lire le 1er caractre de la ligne suivante */
        car = getchar ();
    }
}



/* rcupration des valeurs numriques sur lesquelles portera le tri */
recupnum ()
{
    long  positlig;         /* position ligne en cours d'analyse */
    octet signe;            /* dtection de valeurs ngatives */
    octet bit32;            /* dtection de valeurs positives > 2^31 */
    int   longueur;         /* longueur de la ligne  analyser */
    int   posit;            /* position dans la ligne */
    octet huge *carligne;   /* caractre courant dans la ligne */
    int   valnegative;      /* signe de la valeur qu'on rcupre */
    long  valeur;           /* valeur numrique rcupre pour le tri */

    /* initialisations */
    positlig = 0;
    signe = 0;
    bit32 = 0;

    while ((positlig >> 20) == 0)
    {
        /* rcuprer la longueur de la ligne */
        memcpy (&longueur, (void *)(memoire + positlig + 3), 2);

        /* limiter la zone de recherche */
        if (posfin < longueur)
            longueur = posfin + 1;

        /* se positionner sur le 1er caractre  analyser */
        carligne = memoire + positlig + 9 + posdeb;
        posit = posdeb;

        /* initialisation */
        valeur = 0;
        valnegative = 0;

        /* sauter les blancs avant la valeur numrique */
        while (posit < longueur && (*carligne == ' ' || *carligne == '\t'))
        {
            /* passage au caractre suivant */
            posit ++;
            carligne ++;
        }

        /* prise en compte du signe */
        if (*carligne == '-' || *carligne == '+')
        {
            /* mmoriser le signe */
            if (*carligne == '-')
            {
                valnegative = 1;
                signe = 1;
            }

            /* passer au caractre suivant */
            posit ++;
            carligne ++;
        }

        /* controle syntaxique */
        if (*carligne < '0' || *carligne > '9' || posit >= longueur)
        {
            fputs ("Prsence d'une ligne sans cl numrique malgr l'option -n\n",
                                                                stderr);
            exit (-4);
        }

        /* rcupration de la valeur numrique */
        while (posit < longueur && '0' <= *carligne && *carligne <= '9')
        {
            if (((unsigned long) valeur < 429496729) ||
                (valeur == 429496729 && *carligne <= '5'))
                valeur = (valeur * 10) + (*carligne & 0x0F);
            else
                debordement ();

            /* passer au caractre suivant */
            posit ++;
            carligne ++;
        }

        /* test de l'utilisation du bit 32 pour la valeur absolue */
        if (valeur & 0x80000000)
            bit32 = 1;

        /* prise en compte du signe */
        if (valnegative)
        {
            if (bit32)
                debordement ();
            else
                valeur = -valeur;
        }

        /* recopier la valeur numrique dans la ligne */
        *(memoire + positlig + 5) = (octet)(valeur >> 24);
        *(memoire + positlig + 6) = (octet)(valeur >> 16);
        *(memoire + positlig + 7) = (octet)(valeur >> 8);
        *(memoire + positlig + 8) = (octet) valeur;
        memcpy (&positlig, (void *)(memoire + positlig), 3);
    }

    /* traitement complmentaire pour les valeurs signes */
    if (signe)
    {
        /* test de d'absence de dbordement non dtect */
        if (bit32)
            debordement ();

        /* on se repositionne sur la premire ligne */
        positlig = 0;

        /* et pour chaque ligne */
        while ((positlig >> 20) == 0)
        {
            /* on inverse le bit de signe */
            *(memoire + positlig + 5) = *(memoire + positlig + 5) ^ 0x80;

            /* avant de passer  la ligne suivante */
            memcpy (&positlig, (void *)(memoire + positlig), 3);
        }
    }

    /* la recherche se fera sur la cl numrique */
    posdeb = 0;
    posfin = 3;
}



/* terminaison sur dbordement numrique d'un tri avec l'option -n */
debordement ()
{
    fprintf (stderr, "%s\n%s %lu\n%s %ld  +%ld\n",
             "Dbordement numrique", "Intervalle autoris : de 0 ",
              0xFFFFFFFF, "        ou de", 0x80000001, 0x7FFFFFFF);

    exit (-4);
}



/* tri alphabtique d'une liste de lignes  partir du positime caractre */
tri (long posmem, long postables, int posit)
{
    /* on va dcouper la liste de lignes reue en paramtre en autant de sous
       listes qu'il y aura de valeur diffrentes pour le positime caractre */

    octet far *sousliste;    /* sous listes de lignes */
    octet far *occurence;    /* taille de chaque sous liste */
    long posligcour;         /* mmorisation position de la ligne courante */
    int longueur;            /* longueur de la ligne courante */
    int i, nbpleines;        /* compteurs */
    octet carligne;          /* caractre test, choix de la sous liste */
    long  listetriee;        /* sous liste des lignes entirement explores */


    /* test de dbordement mmoire */
    if (tailletables - postables < 256)
    {
        fputs ("Manque de mmoire lors du tri, utiliser gsort\n", stderr);
        exit (-3);
    }

    /* initialisation des adresses des tables */
    sousliste = (void *)(souslistes + (postables * 3));
    occurence = (void *)(occurences + postables);

    /* et de leur contenu */
    memset (sousliste, 0xFF, 256 * 3);
    memset (occurence, 0, 256);

    /* filtrage de l'octet de poids fort */
    posmem = posmem & 0xFFFFFF;

    /* tant qu'il y a des lignes  analyser */
    while ((posmem >> 20) == 0)
    {
        /* rcuprer la longueur de la ligne  tester */
        memcpy (&longueur, (void *)(memoire + posmem + 3), 2);

        /* si la ligne est assez longue pour contenir le caractre  tester */
        if (posit < longueur)
        {
            /* rcuprer ce caractre */
            carligne = valcar (*(memoire + posmem + 5 + posit));

            /* prise en compte de l'option -r + retournement de la liste */
            if (optr)
                carligne = carligne - 1 + optn;
            else
                carligne = 255 - carligne;
        }
        /* sinon, la ligne sera classe avant toute ligne
           contenant un caractre  la position teste */
        else
            carligne = 255;

        /* rcuprer l'adresse de la ligne suivante */
        posligcour = posmem & 0xFFFFFF;
        memcpy (&posmem, (void *)(memoire + posmem), 3);

        /* complter la sous liste */
        memcpy ((void *)(memoire + posligcour), sousliste + (3 * carligne), 3);

        /* dplacer la ligne teste dans la bonne sous liste */
        memcpy (sousliste + (carligne * 3), &posligcour, 3);

        /* on compte les premires occurences de chaque sous liste */
        if (*(occurence + carligne) <= 2)
            *(occurence + carligne) = *(occurence + carligne) + 1;
    }

    /* rcuprer la liste des lignes totalement explores (tri alphabtique) */
    listetriee = adrliste (sousliste, 255);

    /* si tri alphabtique sans l'option -r, commencer
       par afficher les lignes entirement explores */
    if ((optr | optn) == 0)
        edition (listetriee);

    /* suppression des sous listes vides */
    nbpleines = 0;

    for (i = 0; i < 255 + optn; i++)
    {
        if (*(occurence + i))
        {
            *(occurence + nbpleines) = *(occurence + i);
            memcpy (sousliste + (nbpleines * 3), sousliste + (i * 3), 3);
            nbpleines ++;
        }
    }

    /* pour chaque sous liste restante */
    for (i = nbpleines - 1; i >= 0; i--)
    {
        /* traitement en fonction du nombre de lignes de la sous liste */
        switch (*(occurence + i))
        {
                      /* sous liste vide, on n'a rien  faire */
            case 0  : break;

                      /* sous liste d'une seule ligne, on va l'afficher */
            case 1  : editligne (adrliste (sousliste, i));
                      break;

                      /* sous liste de 2 lignes, on fait un tri simplifi */
            case 2  : tricouple (adrliste (sousliste, i), posit + 1);
                      break;

                      /* au moins 3 lignes, on continue le tri si la
                         zone de tri n'a pas t entirement explore */
            default : if (posit < posfin)
                          tri (adrliste (sousliste, i), postables + nbpleines, posit + 1);
                      else
                          edition (adrliste (sousliste, i));
        }
    }

    /* si tri alphabtique avec option -r, terminer
       en affichant les lignes entirement explores */
    if (optr > optn)
        edition (listetriee);
}



/* retourne l'adresse de la premire ligne d'une sous liste */
long adrliste (octet huge *adrtable, int index)
{
    long retour;

    /* rcupration de l'adresse de la ligne dans la sousliste */
    memcpy (&retour, (void *)(adrtable + (3 * index)), 3);

    /* valeur retourne filtre sur 3 octets */
    return (retour & 0xFFFFFF);
}



/* tri alphabtique simplifi de 2 lignes  partir du positime caractre */
tricouple (long positlig, int posit)
{
    long posligsuiv;                    /* adresse de la 2me ligne */
    int  longueur1, longueur2;          /* longueur de chaque ligne */
    octet huge *carlig1, huge *carlig2; /* caractre courant dans chaque ligne */
    int  sens;                          /* rsultat de la comparaison */

    /* filtrage de l'octet de poids fort de positlig */
    positlig = positlig & 0xFFFFFF;

    /* rcuprer l'adresse de la 2me ligne */
    posligsuiv = 0;
    memcpy (&posligsuiv, (void *)(memoire + positlig), 3);

    /* rcuprer la longueur de chaque ligne */
    memcpy (&longueur1, (void *)(memoire + positlig + 3), 2);
    memcpy (&longueur2, (void *)(memoire + posligsuiv + 3), 2);

    /* se positionner sur le 1er caractre  tester */
    carlig1 = memoire + positlig + 5 + posit;
    carlig2 = memoire + posligsuiv + 5 + posit;

    /* tant que pas de diffrence trouve, pas de fin
       de ligne et zone de tri non entirement explore */
    while (posit < posfin && posit < longueur1
        && posit < longueur2 && valcar (*carlig1) == valcar (*carlig2))
    {
        /* passer aux caractres suivants */
        carlig1 ++;
        carlig2 ++;
        posit ++;
    }

    /* test de la cause de sortie de la boucle */
    if (posit < longueur1 && posit < longueur2)
        /* l'ordre d'affichage dpend du 1er caractre diffrent trouv */
        sens = valcar (*carlig1) - valcar (*carlig2);
    else
        /* l'ordre d'affichage dpend de la longueur des lignes */
        sens = longueur1 - longueur2;

    /* prise en compte de l'option -r */
    if (optr)
        sens = -sens;

    /* afficher les 2 lignes dans le bon ordre */
    if (sens <= 0)
    {
        editligne (positlig);
        editligne (posligsuiv);
    }
    else
    {
        editligne (posligsuiv);
        editligne (positlig);
    }
}



/* retourne le code associ  un caractre */
octet valcar (octet caract)
{
    /* tables pour suppression des accents sous MS-DOS et Windows */
    static octet tabledos [] = "CueaaaaceeeiiiAAEooouuyOUaiounN";
    static octet tablewin [] =
      "AAAAAACEEEEIIIIDNOOOOOUUUUYaaaaaaceeeeiiiionooooouuuuyy";

    octet valcaract;  /* code associ au caractre caract */

    /* initialisation */
    valcaract = caract;

    /* si tri alphabtique */
    if (! optn)
    {
        /* si caractre compris entre ^@ et ^I (code ASCII de 0  9) */
        if (valcaract < '\n')
            /* incrmenter le code de 1 */
            valcaract ++;

        /* si option -d et caractre MS-DOS accentu */
        else if (valcaract >= 0x80 && valcaract <= 0xA5 && optd)
            /* supprimer l'accent */
            valcaract = tabledos [valcaract & 0x3F];

        /* si option -w et caractre Windows accentu */
        else if (valcaract >= 0xC0 && optw)
            /* supprimer l'accent */
            valcaract = tablewin [valcaract & 0x3F];
    }

    /* retourner le code obtenu */
    return valcaract;
}



/* affichage d'une liste chaine de lignes */
edition (long posmem)
{
    /* filtrage de l'octet de poids fort */
    posmem = posmem & 0xFFFFFF;

    /* tant qu'il reste des lignes  afficher */
    while ((posmem >> 20) == 0)
    {
        /* afficher la ligne courante */
        editligne (posmem);

        /* passer  la ligne suivante dans la liste chaine */
        memcpy (&posmem, (void *)(memoire + posmem), 3);
    }
}



/* affichage d'une ligne */
editligne (long posmem)
{
    int longueur;           /* longueur de la ligne  afficher */
    octet huge *carchaine;  /* caractre  afficher */
    int i;                  /* compteur */

    /* rcuprer la longueur de la ligne */
    memcpy (&longueur, (void *)(memoire + posmem + 3), 2);

    /* se positionner sur le 1er caractre de la ligne */
    carchaine = memoire + posmem + posdebl;

    /* afficher chaque caractre de la ligne */
    for (i = 0; i < longueur; i++)
    {
        putchar (*carchaine);
        carchaine ++;
    }

    /* et passer  la ligne suivante de l'cran */
    putchar ('\n');
}
